/*
 * DESCRIPTION:
 * Camera code.
*/

#include "camera.h"

/* Initialize perspective camera. */
void
camera_perspective_init(struct Camera* self, f32 fov, f32 aspect)
{
	memset(self, '\0', sizeof(struct Camera));
	
	self->type = CAMERA_PERSPECTIVE;

	self->fov = fov;
	self->yaw = CAMERA_PERSPECTIVE_YAW_DEFAULT;
	self->pitch = 0;

	self->aspect = aspect;

	camera_update(self);
}

/* Initialize orthographic camera. */
void
camera_orthographic_init(struct Camera* self, vec2 size)
{
	memset(self, '\0', sizeof(struct Camera));

	self->type = CAMERA_ORTHOGRAPHIC;
	glm_vec2_copy(size, self->size);

	camera_update(self);
}

/* Returns view matrix of camera. */
void
camera_view_set(struct Camera* self)
{
	vec3 positionFront;

	glm_mat4_identity(self->view);

	switch (self->type)
	{
		case CAMERA_PERSPECTIVE:
			glm_vec3_add(self->position, self->front, positionFront);
			glm_lookat(self->position, positionFront, self->up, self->view);
			break;
		case CAMERA_ORTHOGRAPHIC:
			glm_mat4_identity(self->view);
			break;
		default:
			break;
	}
}

/* Gets projection of camera. */
void
camera_projection_set(struct Camera* self)
{
	glm_mat4_identity(self->projection);

	switch (self->type)
	{
		case CAMERA_PERSPECTIVE:
			glm_perspective
			(
			 	self->fov,
				self->aspect,
				CAMERA_PERSPECTIVE_NEAR,
				CAMERA_PERSPECTIVE_FAR,
				self->projection
			);
			break;
		case CAMERA_ORTHOGRAPHIC:
			glm_ortho
			(
				0.0f,
				self->size[0], 
				self->size[1], 
				0.0f, 
				CAMERA_ORTHOGRAPHIC_NEAR, 
				CAMERA_ORTHOGRAPHIC_FAR,
				self->projection
			);
			break;
		default:
			break;
	}
}

/* Sets the camera to a position. */
void
camera_position_set(struct Camera* self, vec3 position)
{
	glm_vec3_copy(position, self->position);
}

/* Sets camera's vectors. */
void
camera_vector_set(struct Camera* self)
{
	vec3 worldUp = {0.0f, 1.0f, 0.0f};

	switch (self->type)
	{
		case CAMERA_PERSPECTIVE:

			self->front[0] = cos(self->yaw) * cos(self->pitch);
			self->front[1] = sin(self->pitch);
			self->front[2] = sin(self->yaw) * cos(self->pitch);
			glm_vec3_normalize(self->front);
			glm_vec3_crossn(self->front, worldUp, self->right);
			glm_vec3_crossn(self->right, self->front, self->up);
			break;
		case CAMERA_ORTHOGRAPHIC:
			break;
		default:
			break;
	}
}

/* Sets an orthographic camera's size. */
void
camera_orthographic_size_set(struct Camera* self, vec2 size)
{
	glm_vec2_copy(size, self->size);
}

/* Sets an orthographic camera's size to a window's. */
void
camera_orthographic_window_set(struct Camera* self, struct Window* window)
{
	camera_orthographic_size_set(self, window->size);
}

/* Updates camera. */
void
camera_update(struct Camera* self)
{
	camera_vector_set(self);
	camera_view_set(self);
	camera_projection_set(self);
}
