/*
 * DESCRIPTION:
 * OpenGL shader abstraction.
*/

#include "shader.h"

/* Given two GLSL strings representing the vertex/fragment shaders,
 * initializes and compiles shader program. */
bool
shader_init(struct Shader* self, char* vertex, char* fragment)
{
	s32 success;
	char log[SHADER_LOG_MAX];

	GLuint vertexHandle;
	GLuint fragmentHandle;

	vertexHandle = glCreateShader(GL_VERTEX_SHADER);
	fragmentHandle = glCreateShader(GL_FRAGMENT_SHADER);

	glShaderSource(vertexHandle, 1, (const char* const*)&vertex, NULL);
	glShaderSource(fragmentHandle, 1, (const char* const*)&fragment, NULL);

	glCompileShader(vertexHandle);
	glCompileShader(fragmentHandle);

	glGetShaderiv(vertexHandle, GL_COMPILE_STATUS, &success);

	if (!success)
	{
		glGetShaderInfoLog(vertexHandle, SHADER_LOG_MAX, NULL, log);
		printf("COMPILE ERROR: Shader #%i | %s\n", vertexHandle, log);
		return false;
	}
	else
		printf("COMPILE SUCCESS: Shader #%i\n", vertexHandle);

	glGetShaderiv(fragmentHandle, GL_COMPILE_STATUS, &success);

	if (!success)
	{
		glGetShaderInfoLog(fragmentHandle, SHADER_LOG_MAX, NULL, log);
		printf("COMPILE ERROR: Shader #%i | %s\n", fragmentHandle, log);
		return false;
	}
	else
		printf("COMPILE SUCCESS: Shader #%i\n", fragmentHandle);

	self->handle = glCreateProgram();

	glAttachShader(self->handle, vertexHandle);
	glAttachShader(self->handle, fragmentHandle);

	glLinkProgram(self->handle);

	glDeleteShader(vertexHandle);
	glDeleteShader(fragmentHandle);

	self->isInit = true;

	return true;
}

/* Uses the specified shader program. */
void
shader_use(struct Shader* self)
{
	glUseProgram(self->handle);
}

/* Frees a shader program. */
void
shader_free(struct Shader* self)
{
	if (self->isInit)
	{
		glDeleteProgram(self->handle);
		self->isInit = false;
	}
}

/* Sets a s32 uniform. */
void
shader_uniform_s32_set(struct Shader* self, char* name, s32 value) 
{
	glUniform1i(glGetUniformLocation(self->handle, name), value); 
}

/* Sets a f32 uniform. */
void
shader_uniform_f32_set(struct Shader* self, char* name, f32 value) 
{ 
	glUniform1i(glGetUniformLocation(self->handle, name), value); 
}

/* Sets a vec3 uniform. */
void
shader_uniform_vec3_set(struct Shader* self, char* name, vec3 value)
{
	glUniform3fv(glGetUniformLocation(self->handle, name), 1, value);
}

/* Sets a vec4 uniform. */
void
shader_uniform_vec4_set(struct Shader* self, char* name, vec4 value)
{
	glUniform4fv(glGetUniformLocation(self->handle, name), 1, value);
}

/* Sets a mat4 uniform. */
void
shader_uniform_mat4_set(struct Shader* self, char* name, mat4 value)
{
	glUniformMatrix4fv(glGetUniformLocation(self->handle, name), 1, GL_FALSE, (f32*)value);
}

/* Sets a texture uniform. */
void
shader_uniform_texture_set(struct Shader* self, char* name, struct Texture* texture, s32 index)
{
	glActiveTexture(GL_TEXTURE0 + index);
	texture_bind(texture);
	glUniform1i(glGetUniformLocation(self->handle, name), index);
}
