/*
 * DESCRIPTION:
 * Color collide delete component.
 * For when one of the player's colored orbs hits an enemy.
*/

#include "c_color_collide_delete.h"

static void delete_effect(struct CColorCollideDelete* self, u32 id);
static void delete_check(struct CColorCollideDelete* self, u32 id);
static void update_f(struct CColorCollideDelete* self);

/* Visual effect of color collide delete. */
static void
delete_effect(struct CColorCollideDelete* self, u32 id)
{
	struct CColor* color;
	struct CSprite* sprite;
	struct CColorChange* colorChange;
	struct CScale* scale;
	struct CPhysics* collidePhysics;
	struct CParticleSpawn* particleSpawn;
	u32 particleID;
	vec3 position;

	sprite = ecs_get(self->id, ECS_C_SPRITE);
	colorChange = ecs_get(self->id, ECS_C_COLOR_CHANGE);
	color = ecs_get(self->id, ECS_C_COLOR);
	scale = ecs_get(self->id, ECS_C_SCALE);
	collidePhysics = ecs_get(id, ECS_C_PHYSICS);
	particleSpawn = ecs_get(id, ECS_C_PARTICLE_SPAWN);

	if (!sprite || !colorChange || !scale || !color || !collidePhysics || !particleSpawn)
		return;

	c_color_change_set
	(
		colorChange, 
		C_COLOR_VALUES[color->type],
		COLOR_OPAQUE,
		C_COLOR_COLLIDE_DELETE_COLOR_CHANGE_TIMER_MAX, 
		C_COLOR_CHANGE_FROM
	);

	c_scale_set
	(
	 	scale, 
		C_COLOR_COLLIDE_DELETE_SCALE_BASE_VALUE, 
		C_COLOR_COLLIDE_DELETE_SCALE_VALUE, 
		C_COLOR_COLLIDE_DELETE_SCALE_TIMER_MAX, 
		C_COLOR_COLLIDE_DELETE_SCALE_TYPE
	);

	glm_vec3_copy(collidePhysics->position, position);

	c_particle_spawn(particleSpawn, position);
	
	sound_play(&game.sounds[SOUND_COLLIDE], -1);
}

/* 
 * Checks if the color of the collision is the same as this.
 * If so, removes it.
*/
static void
delete_check(struct CColorCollideDelete* self, u32 id)
{
	struct CColor* color;
	struct CColor* collideColor;

	color = ecs_get(self->id, ECS_C_COLOR);
	collideColor = ecs_get(id, ECS_C_COLOR);

	if (!color || !collideColor)
		return;

	if (collideColor->type == color->type)
	{
		delete_effect(self, id);
		self->isCollided = true;
		self->collideID = id;
		entity_delete(id);
	}
}

/* Draws a color collide delete component. */
static void
update_f(struct CColorCollideDelete* self)
{
	struct CCircleCollide* circleCollide;

	if (self->isDisabled)
		return;

	circleCollide = ecs_get(self->id, ECS_C_CIRCLE_COLLIDE);

	self->isCollided = false;

	if (!circleCollide)
		return;

	if (circleCollide->collisions.count == 0)
		return;

	for (s32 i = 0; i < (s32)circleCollide->collisions.count; i++)
	{
		u32* id;

		id = (u32*)vector_get(&circleCollide->collisions, i);

		delete_check(self, *id);
	}
}

/* Registers color collide delete component. */
void
c_color_collide_delete_register(void)
{
	ecs_register
	(
		ECS_C_COLOR_COLLIDE_DELETE,
		sizeof(struct CColorCollideDelete),
		C_COLOR_COLLIDE_DELETE_LIMIT,
		NULL,
		NULL,
		(ECSFunction)update_f,
		NULL,
		NULL
	);
}
