/*
 * DESCRIPTION:
 * Orbit component; for the follower orbs of the player.
*/

#include "c_orbit.h"

static void object_position_set(struct COrbit* self, u32 id, s32 index);

static void add(struct COrbit* self);
static void delete(struct COrbit* self);
static void update_f(struct COrbit* self);

/* Sets the position of a object in orbit. */
static void
object_position_set(struct COrbit* self, u32 id, s32 index)
{
	struct CPhysics* physics;
	struct CRotation* rotation;
	struct CPhysics* objectPhysics;
	f32 offset;
	vec3 position;

	physics = ecs_get(self->id, ECS_C_PHYSICS);
	rotation = ecs_get(self->id, ECS_C_ROTATION);
	objectPhysics = ecs_get(id, ECS_C_PHYSICS);

	if (!physics || !rotation || !objectPhysics)
		return;

	offset = (f32)(RADIANS_MAX / self->objects.count) * index;

	glm_vec3_copy(physics->position, position);

	position[0] -= physics->velocity[0];
	position[1] -= physics->velocity[1];

	objectPhysics->position[0] = position[0] + (-cos(rotation->angle + offset)) * self->offset;
	objectPhysics->position[1] = position[1] + (sin(rotation->angle + offset)) * self->offset;
}

/* Adds orbit component. */
static void
add(struct COrbit* self)
{
	vector_init(&self->objects, sizeof(u32), C_ORBIT_GAME_OBJECT_LIMIT);
}

/* Deletes orbit component. */
static void
delete(struct COrbit* self)
{
	for (s32 i = 0; i < (s32)self->objects.count; i++)
	{
		u32* id;

		id = (u32*)vector_get(&self->objects, i);

		entity_delete(*id);
	}

	vector_free(&self->objects);
}

/* Draws a orbit component. */
static void
update_f(struct COrbit* self)
{
	if (self->objects.count == 0)
		return;
	
	for (s32 i = 0; i < (s32)self->objects.count; i++)
	{
		u32* id;

		id = (u32*)vector_get(&self->objects, i);
	
		object_position_set(self, *id, i);
	}
}

/* Given a GameObject id, adds it to the orbit. */
void
c_orbit_object_add(struct COrbit* self, u32 id)
{
	vector_push(&self->objects, &id);
}

/* Given a GameObjct id, removes it from orbit. */
void
c_orbit_object_delete(struct COrbit* self, u32 id)
{
	vector_remove_from_data(&self->objects, &id);
}

/* Removes all objects from COrbit. */
void
c_orbit_object_clear(struct COrbit* self, u32 id)
{
	vector_clear(&self->objects);
}

/* Sets orbit component. */
void
c_orbit_set(struct COrbit* self, f32 offset)
{
	self->offset = offset;
}

/* Registers orbit component. */
void
c_orbit_register(void)
{
	ecs_register
	(
		ECS_C_ORBIT,
		sizeof(struct COrbit),
		C_ORBIT_LIMIT,
		(ECSFunction)add,
		(ECSFunction)delete,
		(ECSFunction)update_f,
		NULL,
		NULL
	);
}
