/*
 * DESCRIPTION:
 * Player death component.
*/

#include "c_player_death.h"

static void state_set(struct CPlayerDeath* self, enum CPlayerDeathState state);
static void add(struct CPlayerDeath* self);
static void begin(struct CPlayerDeath* self);
static void update_f(struct CPlayerDeath* self);

/* Sets death state. */
static void
state_set(struct CPlayerDeath* self, enum CPlayerDeathState state)
{
	struct CScale* scale;
	struct CColorChange* colorChange;

	scale = ecs_get(self->id, ECS_C_SCALE);
	colorChange = ecs_get(self->id, ECS_C_COLOR_CHANGE);

	self->state = state;

	switch (self->state)
	{
		case C_PLAYER_DEATH_PULSATE:
			self->timer = C_PLAYER_DEATH_PULSATE_TIME;
			break;
		case C_PLAYER_DEATH_GROW:
			c_scale_set
			(
				scale, 
				C_PLAYER_DEATH_GROW_SCALE_BASE_VALUE, 
				C_PLAYER_DEATH_GROW_SCALE_VALUE, 
				C_PLAYER_DEATH_GROW_TIME, 
				C_SCALE_TO_STICK
			);
			
			c_color_change_set
			(
				colorChange, 
				C_PLAYER_DEATH_GROW_COLOR_CHANGE_BASE_COLOR, 
				C_PLAYER_DEATH_GROW_COLOR_CHANGE_COLOR,
				C_PLAYER_DEATH_GROW_TIME,
				C_COLOR_CHANGE_TO_STICK
			);

			self->timer = C_PLAYER_DEATH_GROW_TIME;
			break;
		case C_PLAYER_DEATH_VANISH:
			c_scale_set
			(
				scale, 
				C_PLAYER_DEATH_VANISH_SCALE_BASE_VALUE, 
				C_PLAYER_DEATH_VANISH_SCALE_VALUE, 
				C_PLAYER_DEATH_VANISH_TIME, 
				C_SCALE_TO_STICK
			);
			
			c_color_change_set
			(
				colorChange, 
				C_PLAYER_DEATH_VANISH_COLOR_CHANGE_BASE_COLOR, 
				C_PLAYER_DEATH_VANISH_COLOR_CHANGE_COLOR,
				C_PLAYER_DEATH_VANISH_TIME,
				C_COLOR_CHANGE_TO_STICK
			);

			self->timer = C_PLAYER_DEATH_VANISH_TIME;
			break;
		case C_PLAYER_DEATH_NONE:
		case C_PLAYER_DEATH_END:
		default:
			break;
	}
}

/* Begins death process. */
static void
begin(struct CPlayerDeath* self)
{
	struct CRotation* rotation;
	struct CControl* control;
	struct COrbit* orbit;
	struct CPulsate* pulsate;

	control = ecs_get(self->id, ECS_C_CONTROL);
	rotation = ecs_get(self->id, ECS_C_ROTATION);
	orbit = ecs_get(self->id, ECS_C_ORBIT);
	pulsate = ecs_get(self->id, ECS_C_PULSATE);
	
	if (!control || !rotation || !orbit)
		return;

	control->isDisabled = true;
	rotation->isDisabled = true;
	pulsate->isDisabled = true;

	for (s32 i = 0; i < (s32)orbit->objects.count; i++)
	{
		struct CTimeDelete* timeDelete;
		struct CScale* scale;
		struct CColorChange* colorChange;
		struct CColorCollideDelete* colorCollideDelete;
		struct CSprite* sprite;
		u32* id;

		id = (u32*)vector_get(&orbit->objects, i);

		timeDelete = ecs_get(*id, ECS_C_TIME_DELETE);
		scale = ecs_get(*id, ECS_C_SCALE);
		colorChange = ecs_get(*id, ECS_C_COLOR_CHANGE);
		sprite = ecs_get(*id, ECS_C_SPRITE);
		colorCollideDelete = ecs_get(*id, ECS_C_COLOR_COLLIDE_DELETE);

		if (!timeDelete || !scale || !colorChange || !sprite || !colorCollideDelete)
			continue;

		colorCollideDelete->isDisabled = true;

		c_scale_set
		(
			scale, 
			C_PLAYER_DEATH_FOLLOWER_BASE_SCALE,
			C_PLAYER_DEATH_FOLLOWER_SCALE, 
			C_PLAYER_DEATH_FOLLOWER_TIMER, 
			C_SCALE_TO
		);

		c_color_change_set
		(
			colorChange, 
			sprite->color, 
			C_PLAYER_DEATH_FOLLOWER_COLOR, 
			C_PLAYER_DEATH_FOLLOWER_TIMER,
			C_COLOR_CHANGE_TO
		);

		c_time_delete_set(timeDelete, C_PLAYER_DEATH_FOLLOWER_TIMER);
	}

	state_set(self, C_PLAYER_DEATH_PULSATE);
}

/* Adds a CPlayerDeath component. */
static void
add(struct CPlayerDeath* self)
{
	state_set(self, C_PLAYER_DEATH_NONE);
}

/* Draws a player_death component. */
static void
update_f(struct CPlayerDeath* self)
{
	struct CHealth* health;

	health = ecs_get(self->id, ECS_C_HEALTH);

	if (!health || !health->isDead || self->state == C_PLAYER_DEATH_END)
		return;

	if (self->state == C_PLAYER_DEATH_NONE)
		begin(self);

	self->timer--;

	if (self->timer <= 0)
		state_set(self, self->state + 1);
}

/* Registers player_death component. */
void
c_player_death_register(void)
{
	ecs_register
	(
		ECS_C_PLAYER_DEATH,
		sizeof(struct CPlayerDeath),
		C_PLAYER_DEATH_LIMIT,
		(ECSFunction)add,
		NULL,
		(ECSFunction)update_f,
		NULL,
		NULL
	);
}
