/*
 * DESCRIPTION:
 * Sprite component.
*/

#include "c_sprite.h"

static void c_sprite_size_get(struct CSprite* self, vec2 size);

static void add(struct CSprite* self);
static void draw(struct CSprite* self);
static void update_f(struct CSprite* self);

/* Obtains sprite draw size. */
static void
c_sprite_size_get(struct CSprite* self, vec2 size)
{
	glm_vec2_copy(self->size, size);
	glm_vec2_scale(size, self->scale, size);
}

/* Adds sprite component. */
static void
add(struct CSprite* self)
{
	glm_vec4_copy(COLOR_OPAQUE, self->color);
	self->scale = C_SPRITE_SCALE_DEFAULT;
}

/* Updates (fixed) sprite component. */
static void
update_f(struct CSprite* self)
{
	self->rotation = self->rotation > RADIANS_MAX ? RADIANS_MIN : self->rotation;
	self->rotation = self->rotation < RADIANS_MIN ? RADIANS_MAX : self->rotation;
}

/* Draws a sprite component. */
static void
draw(struct CSprite* self)
{
	mat4 model;
	vec2 size;
	vec2 origin;
	vec3 position;
	vec3 modelOrigin;
	
	c_sprite_size_get(self, size);
	origin_2d_get(self->origin, size, origin);

	position[0] = self->position[0] - origin[0];
	position[1] = self->position[1] - origin[1];
	position[2] = self->position[2];

	modelOrigin[0] = origin[0];
	modelOrigin[1] = origin[1];
	modelOrigin[2] = 0.0f;

	glm_mat4_identity(model);
	glm_translate(model, position);
	glm_rotate_at(model, modelOrigin, self->rotation, C_SPRITE_ROTATION_AXIS);

	shader_texture_quad_draw(&self->atlas, self->shader, model, size, self->color);
}

/* Returns the sprite's containing rectangle. */
void
c_sprite_rectangle_get(struct CSprite* self, vec4 rectangle)
{
	vec2 size;
	vec2 origin;
	
	c_sprite_size_get(self, size);
	origin_2d_get(self->origin, size, origin);

	rectangle[0] = self->position[0] - origin[0];
	rectangle[1] = self->position[1] - origin[1];
	rectangle[2] = rectangle[0] + self->size[0];
	rectangle[3] = rectangle[1] + self->size[1];
}

/* Sets sprite component info. */
void
c_sprite_set
(
	struct CSprite* self, 
	struct Texture texture, 
	enum ShaderType shader, 
	enum OriginType origin, 
	vec2 size, 
	vec4 color, 
	vec3 position
)
{
	c_sprite_atlas_set
	(
		self, 
		texture, 
		shader, 
		origin,
		texture.w, 
		texture.h, 
		1, 
		1, 
		size, 
		color, 
		position
	);
}

/* Sets sprite component info, additionally given atlas rows/columns. */
void
c_sprite_atlas_set
(
	struct CSprite* self, 
	struct Texture texture, 
	enum ShaderType shader,
	enum OriginType origin,
	s32 frameW, 
	s32 frameH, 
	s32 rows, 
	s32 cols, 
	vec2 size, 
	vec4 color, 
	vec3 position
)
{
	self->shader = shader;
	self->origin = origin;

	atlas_init(&self->atlas, texture, frameW, frameH, rows, cols);

	glm_vec2_copy(size, self->size);
	glm_vec3_copy(position, self->position);
	glm_vec4_copy(color, self->color);
}

/* Registers sprite component. */
void
c_sprite_register(void)
{
	ecs_register
	(
		ECS_C_SPRITE,
		sizeof(struct CSprite),
		C_SPRITE_LIMIT,
		(ECSFunction)add,
		NULL,
		(ECSFunction)update_f,
		NULL,
		(ECSFunction)draw
	);
}
