/*
 * DESCRIPTION:
 * Text component.
*/

#include "c_text.h"

static void c_text_free(struct CText* self);
static void c_text_glyph_draw(struct CText* self, struct Texture, vec3 position);

static void add(struct CText* self);
static void draw(struct CText* self);

/* Obtains the size of the glyph. */
static void
c_text_glyph_size_get(struct CText* self, struct Texture texture, vec2 size)
{
	size[0] = texture.w;
	size[1] = texture.h;
	glm_vec2_scale(size, self->scale, size);
}

/* Frees glyph textures. */
static void
c_text_free(struct CText* self)
{
	for (s32 i = 0; i < self->length; i++)
	{
		struct Texture* texture;

		texture = &self->glyphTextures[i];

		if (texture->isInit)
			texture_free(texture);
	}

	memset(&self->glyphTextures, '\0', sizeof(self->glyphTextures));
}

/* Draws a glyph. */
static void
c_text_glyph_draw(struct CText* self, struct Texture texture, vec3 position)
{
	struct Atlas atlas;
	mat4 model;
	vec2 size;

	c_text_glyph_size_get(self, texture, size);

	glm_mat4_identity(model);
	glm_translate(model, position);

	atlas_init(&atlas, texture, texture.w, texture.h, 1, 1);

	shader_texture_quad_draw(&atlas, self->shader, model, size, self->color);
}

/* Adds text component. */
static void
add(struct CText* self)
{
	glm_vec4_copy(COLOR_OPAQUE, self->color);
	self->scale = C_TEXT_SCALE_DEFAULT;
}

/* Deletes text component. */
static void
delete(struct CText* self)
{
	c_text_free(self);
}

/* Draws a text component. */
static void
draw(struct CText* self)
{
	char glyph;
	char nextGlyph;
	f32 advance;
	f32 lineW;
	f32 offsetX;
	f32 kerning;
	struct GlyphMetrics glyphMetrics;
	vec3 position;
	vec4 rectangle;
	vec2 origin;
	vec2 size;
	f32 lineSkip;
	f32 lineBegin;

	c_text_rectangle_get(self, rectangle);

	size[0] = rectangle[2] - rectangle[0];
	size[1] = rectangle[3] - rectangle[1];

	origin_2d_get(self->origin, size, origin);
	
	position[0] = self->position[0] - origin[0];
	position[1] = self->position[1] - origin[1];
	position[2] = self->position[2];

	offsetX = 0.0f;
	lineW = 0.0f;

	lineBegin = position[0];

	lineSkip = font_line_skip_get(&self->font) * self->scale;

	for (s32 i = 0; i < self->length; i++)
	{
		glyph = self->string[i];
	
		if 
		(
			glyph == '\n' ||
			(glyph == ' ' && (lineW > self->wrap && self->wrap != -1))
		)
		{
			position[0] = lineBegin;
			position[1] += lineSkip;

			lineW = 0.0f;

			if (glyph == '\n')
				continue;
		}

		if (i < self->length)
		{
			nextGlyph = self->string[i + 1];
			kerning = (f32)glyph_kerning_get(&self->font, glyph, nextGlyph);
		}
		else
			kerning = 0.0;

		glyph_metrics_get(&glyphMetrics, &self->font, glyph);

		c_text_glyph_draw(self, self->glyphTextures[i], position);

		advance = glyphMetrics.advance;

		offsetX = advance + kerning;
		offsetX *= self->scale;

		position[0] += offsetX;
		lineW += offsetX;
	}
}

/* Gets a rectangle the size of all glyphs. */
void
c_text_rectangle_get(struct CText* self, vec4 rectangle)
{
	char glyph;
	char nextGlyph;
	f32 lineW;
	f32 offsetX;
	f32 prevY;
	struct GlyphMetrics glyphMetrics;
	vec2 origin;
	vec2 size;
	f32 lineSkip;
	f32 kerning;
	f32 advance;

	if (self->length <= 0 || self->length > 255)
		return;

	lineSkip = font_line_skip_get(&self->font);
	lineSkip *= self->scale;

	glyph_metrics_get(&glyphMetrics, &self->font, self->string[0]);

	lineW = 0.0f;

	size[0] = 0.0f;
	size[1] = (f32)glyphMetrics.maxY - glyphMetrics.minY;

	for (s32 i = 0; i < self->length; i++)
	{
		glyph = self->string[i];

		if 
		(
			glyph == '\n' ||
			(glyph == ' ' && (lineW > self->wrap && self->wrap != -1))
		)
		{
			size[1] += lineSkip;
			lineW = 0.0f;

			if (glyph == '\n')
				continue;
		}

		if (i < self->length)
		{
			nextGlyph = self->string[i + 1];
			kerning = (f32)glyph_kerning_get(&self->font, glyph, nextGlyph);
		}
		else
			kerning = 0.0;

		glyph_metrics_get(&glyphMetrics, &self->font, glyph);

		advance = glyphMetrics.advance;

		offsetX = advance + kerning;
		offsetX *= self->scale;
	
		lineW += offsetX;

		if (lineW > size[0])
			size[0] = lineW;
	}

	origin_2d_get(self->origin, size, origin);

	rectangle[0] = self->position[0] - origin[0];
	rectangle[1] = self->position[1] - origin[1];
	rectangle[2] = rectangle[0] + size[0];
	rectangle[3] = rectangle[1] + size[1];
}

/* Given a string and its length, generates textures for the string. */
void
c_text_string_set(struct CText* self, char* string, s32 length)
{
	c_text_free(self);

	if (length > C_TEXT_STRING_LIMIT)
		return;

	self->length = length;

	memset(self->string, '\0', sizeof(char) * C_TEXT_STRING_LIMIT);
	memcpy(self->string, string, sizeof(char) * self->length);

	for (s32 i = 0; i < self->length; i++)
		glyph_texture_init(&self->glyphTextures[i], &self->font, (char)string[i]);
}

/* Sets text component info. */
void
c_text_set
(
	struct CText* self, 
	enum ShaderType shader, 
	enum OriginType origin,
	struct Font font,
	char* string,
	s32 length,
	s32 wrap,
	vec4 color, 
	vec3 position
)
{
	self->font = font;
	self->shader = shader;
	self->wrap = wrap;
	self->origin = origin;

	glm_vec3_copy(position, self->position);
	glm_vec4_copy(color, self->color);

	c_text_string_set(self, string, length);
}

/* Registers text component. */
void
c_text_register(void)
{
	ecs_register
	(
		ECS_C_TEXT,
		sizeof(struct CText),
		C_TEXT_LIMIT,
		(ECSFunction)add,
		(ECSFunction)delete,
		NULL,
		NULL,
		(ECSFunction)draw
	);
}
