/*
 * DESCRIPTION:
 * Entity component system (ECS) functions.
*/

#include "entity.h"

static void ecs_component_function(struct ECSComponentList* list, ECSFunction function);
static void ecs_function(enum ECSFunctionType type);

/* Uses a function on a component list. */
static void
ecs_component_function(struct ECSComponentList* list, ECSFunction function)
{
	for (s32 i = 0; i < (s32)list->components.count; i++)
	{
		void* component;

		component = vector_get(&list->components, i);

		if (!component)
			return;

		function(component);
	}
}

/* Executes the specified function on all components. */
static void
ecs_function(enum ECSFunctionType type)
{
	for (s32 i = 0; i < ECS_C_COUNT; i++)
	{
		ECSFunction function;
		struct ECSComponentList* list;

		list = &game.ecs.lists[i];

		function = list->system.functions[type];

		if (function)
			ecs_component_function(list, function);
	}
}

/* Registers a component. */
void
ecs_register
(
	enum ECSComponentType type, 
	size_t size, 
	u32 limit,
	ECSFunction add, 
	ECSFunction delete, 
	ECSFunction update_f, 
	ECSFunction update_v,
	ECSFunction draw
)
{
	struct ECSComponentList* list;

	list = &game.ecs.lists[type];

	memset(list, '\0', sizeof(struct ECSComponentList));

	list->type = type;

	list->system.functions[ECS_FUNCTION_ADD] = add;
	list->system.functions[ECS_FUNCTION_DELETE] = delete;
	list->system.functions[ECS_FUNCTION_UPDATE_F] = update_f;
	list->system.functions[ECS_FUNCTION_UPDATE_V] = update_v;
	list->system.functions[ECS_FUNCTION_DRAW] = draw;

	vector_init(&list->components, size, limit);
}

/* Unregisters a component. */
void
ecs_unregister(enum ECSComponentType type)
{
	struct ECSComponentList* list;

	list = &game.ecs.lists[type];

	vector_free(&list->components);

	memset(list, '\0', sizeof(struct ECSComponentList));
}

/* Adds a component to the ECS. */
void*
ecs_add(u32 id, enum ECSComponentType type)
{
	ECSFunction add;
	struct ECSComponentList* list;
	void* component;

	list = &game.ecs.lists[type];

	component = vector_push(&list->components, NULL);

	add = list->system.functions[ECS_FUNCTION_ADD];

	if (add)
		add(component);

	memcpy(component, &id, sizeof(u32));

	return component;
}

/* Deletes a component from an ECS. */
void
ecs_delete(u32 id, enum ECSComponentType type)
{
	ECSFunction delete;
	struct ECSComponentList* list;
	void* component;

	component = ecs_get(id, type);

	if (!component)
		return;

	list = &game.ecs.lists[type];

	delete = list->system.functions[ECS_FUNCTION_DELETE];

	if (delete)
		delete(component);

	vector_remove_from_data(&list->components, component);
}

/* Gets a component based on an ID. */
void*
ecs_get(u32 id, enum ECSComponentType type)
{
	struct ECSComponentList* list;

	list = &game.ecs.lists[type];

	for (s32 i = 0; i < (s32)list->components.count; i++)
	{
		u32 componentID;
		void* component;

		component = vector_get(&list->components, i);

		/* All components have an ID as the first member of their struct. */
		memcpy(&componentID, component, sizeof(u32));

		if (componentID == id)
			return component;
	}

	return NULL;
}

/* Gets component based on an index instead. */
void*
ecs_from_index_get(u32 index, enum ECSComponentType type)
{
	struct ECSComponentList* list;

	list = &game.ecs.lists[type];

	return vector_get(&list->components, index);
}

/* Removes all components from a given system. */
void
ecs_clear(enum ECSComponentType type)
{
	struct ECSComponentList* list;

	list = &game.ecs.lists[type];

	vector_clear(&list->components);
}

/* Clears all components in the ECS (but does not remove the systems). */
void
ecs_clear_all(void)
{
	for (s32 i = 0; i < (s32)ECS_C_COUNT; i++)
		ecs_clear((enum ECSComponentType)i);
}

/* Updates (fixed) ECS. */
void
ecs_update_f(void)
{
	ecs_function(ECS_FUNCTION_UPDATE_F);
}

/* Updates (variable) ECS. */
void
ecs_update_v(void)
{
	ecs_function(ECS_FUNCTION_UPDATE_V);
}

/* Draws ECS. */
void
ecs_draw(void)
{
	ecs_function(ECS_FUNCTION_DRAW);
}
