/*
 * DESCRIPTION:
 * Handles game state.
*/

#include "game_state.h"

static void game_state_change(enum StateType state);
static void game_state_texture_init(struct StateData* stateData);
static void game_state_texture_free(struct StateData* stateData);
static void game_state_sound_init(struct StateData* stateData);
static void game_state_sound_free(struct StateData* stateData);
static void game_state_music_init(struct StateData* stateData);
static void game_state_music_free(struct StateData* stateData);
static void game_state_shader_init(struct StateData* stateData);
static void game_state_shader_free(struct StateData* stateData);
static void game_state_font_init(struct StateData* stateData);
static void game_state_font_free(struct StateData* stateData);

/* Inits state shaders. */
static void
game_state_shader_init(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->shaderCount; i++)
	{
		enum ShaderType type;

		type = stateData->shaders[i];

		game_shader_init(type);
	}
}

/* Frees state shaders. */
static void
game_state_shader_free(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->shaderCount; i++)
	{
		enum ShaderType type;

		type = stateData->shaders[i];

		game_shader_free(type);
	}
}

/* Inits state textures. */
static void
game_state_texture_init(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->textureCount; i++)
	{
		enum TextureType type;

		type = stateData->textures[i];

		game_texture_init(type);
	}
}

/* Frees state textures. */
static void
game_state_texture_free(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->textureCount; i++)
	{
		enum TextureType type;

		type = stateData->textures[i];

		game_texture_free(type);
	}
}

/* Inits state sounds. */
static void
game_state_sound_init(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->soundCount; i++)
	{
		enum SoundType type;

		type = stateData->sounds[i];

		game_sound_init(type);
	}
}

/* Frees state sounds. */
static void
game_state_sound_free(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->soundCount; i++)
	{
		enum SoundType type;

		type = stateData->sounds[i];

		game_sound_free(type);
	}
}

/* Inits state music. */
static void
game_state_music_init(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->musicCount; i++)
	{
		enum MusicType type;

		type = stateData->music[i];

		game_music_init(type);
	}
}

/* Frees state music. */
static void
game_state_music_free(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->musicCount; i++)
	{
		enum MusicType type;

		type = stateData->music[i];

		game_music_free(type);
	}
}

/* Inits state font. */
static void
game_state_font_init(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->fontCount; i++)
	{
		enum FontType type;

		type = stateData->fonts[i];

		game_font_init(type);
	}
}

/* Frees state font. */
static void
game_state_font_free(struct StateData* stateData)
{
	for (s32 i = 0; i < stateData->fontCount; i++)
	{
		enum FontType type;

		type = stateData->fonts[i];

		game_font_free(type);
	}
}

/* Changes state. */
void
game_state_change(enum StateType state)
{
	game_state_free();
	game_state_data_free();
	ecs_clear_all();

	game.state = state;

	game_state_data_init();
	game_state_init();
}

/* Initializes state data. */
void
game_state_data_init(void)
{
	struct StateData stateData;

	stateData = STATE_DATA[game.state];

	game_state_shader_init(&stateData);
	game_state_texture_init(&stateData);
	game_state_sound_init(&stateData);
	game_state_music_init(&stateData);
	game_state_font_init(&stateData);
}

/* Initializes state data. */
void
game_state_data_free(void)
{
	struct StateData stateData;

	stateData = STATE_DATA[game.state];

	game_state_shader_free(&stateData);
	game_state_texture_free(&stateData);
	game_state_sound_free(&stateData);
	game_state_music_free(&stateData);
	game_state_font_free(&stateData);
}

/* Initializes the specified game state. */
void
game_state_init(void)
{
	switch (game.state)
	{
		case STATE_MENU:
			menu_init(&game.menu);
			break;
		case STATE_PLAY:
			play_init(&game.play);
			break;
		case STATE_RESULTS:
			results_init(&game.results);
			break;
		default:
			break;
	}
}

/* Updates game state. */
void
game_state_update_f(void)
{
	switch (game.state)
	{
		case STATE_MENU:
			menu_update_f(&game.menu);

			if (game.menu.isStart)
				game_state_change(STATE_PLAY);

			break;
		case STATE_PLAY:
			play_update_f(&game.play);

			if (game.play.isOver)
			{
				game.score = game.play.score;
				game_state_change(STATE_RESULTS);
			}

			break;
		case STATE_RESULTS:
			results_update_f(&game.results);

			if (game.results.isMenu)
				game_state_change(STATE_MENU);

			break;
		default:
			break;
	}
}

/* Frees game state. */
void
game_state_free(void)
{
	switch (game.state)
	{
		case STATE_MENU:
			menu_free(&game.menu);
			break;
		case STATE_PLAY:
			play_free(&game.play);
			break;
		case STATE_RESULTS:
			results_free(&game.results);
			break;
		default:
			break;
	}
}
