/*
 * DESCRIPTION:
 * Dynamic array/vector implementation.
*/

#include "vector.h"

static void vector_realloc(struct Vector* self);

/* Reallocates the vector to a specified size. */
static void
vector_realloc(struct Vector* self)
{
	if (self->count == 0)
	{
		self->data = NULL;
		return;
	}

	self->data = realloc(self->data, vector_size_get(self));
}

/* Pushes all elements past an index forward (<-). */
static void
vector_forward(struct Vector* self, s32 index)
{
	s32 length;

	length = ((self->count - 1) - index);

	for (s32 i = 0; i < length; i++)
	{
		void* data;
		void* nextData;
		s32 currentIndex;

		currentIndex = index + i;

		data = self->data + (currentIndex * self->itemSize);
		nextData = data + self->itemSize;

		memcpy(data, nextData, self->itemSize);
	}

}

/* Pushes all elements past an index backward (->).*/
static void
vector_backward(struct Vector* self, s32 index)
{
	s32 length;

	length = ((self->count - 1) - index);

	for (s32 i = 0; i < length; i++)
	{
		void* data;
		void* prevData;
		s32 index;

		index = (self->count - 1) - i;

		data = self->data + (index * self->itemSize);
		prevData = data - self->itemSize;

		memcpy(data, prevData, self->itemSize);
	}
}

/* Returns total size of vector. */
size_t
vector_size_get(struct Vector* self)
{
	return self->count * self->itemSize;
}

/* Initializes a vector, given a size of an element. */
void
vector_init(struct Vector* self, size_t size, s32 limit)
{
	memset(self, '\0', sizeof(struct Vector));

	self->itemSize = size;
	self->count = 0;
	self->limit = limit;

	vector_realloc(self);

	memset(self->data, '\0', vector_size_get(self));
}

/* Frees a vector. */
void
vector_free(struct Vector* self)
{
	if (self->data)
		free(self->data);

	self->count = 0;
}

/* Adds an element to the end of the vector. */
void* 
vector_push(struct Vector* self, void* data)
{
	void* p;

	self->count++;

	if (self->count > self->limit)
	{
		self->count = self->limit;
		return NULL;
	}

	vector_realloc(self);

	p = self->data + (vector_size_get(self) - self->itemSize);

	if (data)
		memcpy(p, data, self->itemSize);
	else
		memset(p, '\0', self->itemSize);

	return p;
}

/* Removes the last element of the vector. */
void
vector_pop(struct Vector* self)
{
	self->count--;

	vector_realloc(self);
}

/* Removes all elements of a vector (but does not dealloc the data). */
void
vector_clear(struct Vector* self)
{
	self->count = 0;

	vector_realloc(self);
}

/* Inserts an element at the specified index, putting successive elements back. */
void*
vector_insert(struct Vector* self, void* data, s32 index)
{
	void* p;

	self->count++;

	if (self->count > self->limit)
	{
		self->count = self->limit;
		return NULL;
	}

	vector_realloc(self);

	vector_backward(self, index);

	p = self->data + (index * self->itemSize);

	memcpy(p, data, self->itemSize);

	return p;
}

/* Removes an element in a vector at the specified index, bringing successive elements forward. */
void 
vector_remove(struct Vector* self, s32 index)
{
	vector_forward(self, index);

	vector_pop(self);
}

/* Returns the address of the element at the specified index. */
void*
vector_get(struct Vector* self, s32 index)
{
	return self->data + (index * self->itemSize);
}

/* Returns the index of an element, given it is valid. */
s32
vector_index_get(struct Vector* self, void* data)
{
	s32 index;

	index = (data - (void*)self->data) / self->itemSize;

	return index;
}

/* Removes an element within a vector, given its data, provided it is valid. */
void
vector_remove_from_data(struct Vector* self, void* data)
{
	vector_remove(self, vector_index_get(self, data));
	data = NULL;
}
